// include the basic windows header files and the Direct3D header file
#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <DxErr.h>
#include <math.h>
#include "resource.h"

// define the screen resolution
#define MAX_BUFFER_WIDTH 2000
#define MAX_BUFFER_HEIGHT 2000
#define Q_Round(a) (abs(a - (int)a)>0.5?(int)ceil(a):(int)floor(a))

const int texture_size = 4096;
const int pixel_size = 4;
const int max_file = pixel_size * texture_size*texture_size;

// global declarations
LPDIRECT3D9 d3d = NULL;
LPDIRECT3DDEVICE9 d3ddev = NULL;
LPDIRECT3DVERTEXBUFFER9 v_buffer = NULL;

LPD3DXEFFECT dxEffect;

// function prototypes
bool initD3D(HWND hWnd);
void render_frame(void);
void cleanD3D(void);
bool init_graphics(void);
LRESULT CALLBACK WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
bool write_file_to_texture(const char *file_path, LPDIRECT3DTEXTURE9 texture);
void UpdateWindowSize(HWND hWnd);
void UpdateTitle(HWND hWnd);

struct CUSTOMVERTEX {FLOAT X, Y, Z, RHW; DWORD COLOR; FLOAT U, V; };
#define CUSTOMFVF (D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1)

int window_width = 800, window_height = 600;
int buffer_width = 800, buffer_height = 600;

int data_array_length;
LPDIRECT3DTEXTURE9 texture;
int shader_scale = 1;
bool directx_initialized = false;
bool resize_inprogress = false;
int scroll_offset = 0;
int shader_color = 1;

// the entry point for any Windows program
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    HWND hWnd;
    WNDCLASSEX wc;
	
    ZeroMemory(&wc, sizeof(WNDCLASSEX));
	
    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.hInstance = hInstance;
    wc.hCursor = LoadCursor(NULL, IDC_CROSS);
	wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON));
	wc.hIconSm = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON));
    wc.lpszClassName = "WindowClass";
	
    RegisterClassEx(&wc);
	
    hWnd = CreateWindowEx(WS_EX_ACCEPTFILES, "WindowClass", "Data Viewer", WS_OVERLAPPEDWINDOW,
						  0, 0, 800, 600, NULL, NULL, hInstance, NULL);
	
    ShowWindow(hWnd, nCmdShow);
	
    // set up and initialize Direct3D
	
    bool bRes = initD3D(hWnd);
	if (!bRes)
	{
		return -1;
	}

	if (strlen(lpCmdLine))
	{
		//strip quotation marks
		char fileName[1024];
		strcpy_s<1024>(fileName, &lpCmdLine[1]);
		int idx = (int)(strchr(fileName, '"')-fileName);
		if (idx > 0 && idx < 1024)
			fileName[idx] = '\0';
		write_file_to_texture(fileName, texture);
	}

	
	UpdateWindowSize(hWnd);
	
    // enter the main loop:
    MSG msg;
	
    while(TRUE)
    {
        while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
		
        if(msg.message == WM_QUIT)
            break;
		
        render_frame();
    }
	
    cleanD3D();
	
    return msg.wParam;
}

void UpdateTitle(HWND hWnd)
{
	char Title[256];
	
	int byte_width = buffer_width / shader_scale, byte_height = buffer_height / shader_scale;
	int kbyte_size = (int)ceil((byte_width * byte_height) / 1024.0f);
	sprintf_s<256>(Title,  "Data Viewer (%d x %d) bytes   (%d) KB offset   (%d total) kbytes", byte_width, byte_height, scroll_offset / 1024, kbyte_size);
	
	SetWindowText(hWnd, Title);
}

void UpdateWindowSize(HWND hWnd)
{
	RECT client_size;
	GetClientRect(hWnd, &client_size);
				
	//size must be multiple of scalefactor
	float height = (float)(client_size.bottom - client_size.top), width = (float)(client_size.right - client_size.left);
	buffer_height = Q_Round(height / shader_scale) * shader_scale;
	buffer_width = Q_Round(width / shader_scale) * shader_scale;
				
	int delta_x = buffer_width - (client_size.right - client_size.left);
	int delta_y = buffer_height - (client_size.bottom - client_size.top);

	if (!resize_inprogress && (delta_x != 0 || delta_y != 0))
	{
		resize_inprogress = true;
		RECT position;
		GetWindowRect(hWnd, &position);
					
		int cx = position.right - position.left + (delta_x);
		int cy = position.bottom - position.top + (delta_y);
		
		SetWindowPos(hWnd, NULL, position.left, position.top, cx, cy, 0);
				
		resize_inprogress = false;
	}
				
	RECT window_size;
	GetWindowRect(hWnd, &window_size);
	GetClientRect(hWnd, &client_size);
	
	window_height = window_size.bottom - window_size.top;
	window_width = window_size.right - window_size.left;
	
	UpdateTitle(hWnd);
}

// this is the main message handler for the program
LRESULT CALLBACK WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
    {
        case WM_DESTROY:
        {
            PostQuitMessage(0);
            return 0;
        }
		break;
		case WM_SIZE:
		{
			if (directx_initialized)
				UpdateWindowSize(hWnd);
		}
		break;
		case WM_KEYUP:
		{
			if (wParam == 'G')//set grayscale
			{
				shader_color = 0;
			}
			else if (wParam == 'F')//set falsecolor
			{
				shader_color = 1;
			}
			else if (wParam == 'H')//set High Contrast Falsecolor
			{
				shader_color = 2;
			}
			
			else if (wParam >= '1' && wParam <= '8')//set zoom
			{
				shader_scale = (wParam - (int)'0');
				UpdateWindowSize(hWnd);
			}
		}
		break;
		case WM_DROPFILES:
		{
			HDROP hDrop =  (HDROP)wParam;
			TCHAR szName[MAX_PATH];
			int count = DragQueryFile(hDrop, 0xFFFFFFFF, szName, MAX_PATH);
			
			if (count == 1)
			{
				DragQueryFile(hDrop, 0, szName, MAX_PATH);
				
				write_file_to_texture(szName, texture);
			}

			DragFinish(hDrop);
		}
		break;
		case WM_MOUSEWHEEL:
		{
			short zDelta = GET_WHEEL_DELTA_WPARAM(wParam);
			short accelerate = GET_KEYSTATE_WPARAM(wParam);
			
			if (accelerate & MK_CONTROL)//scroll horizontal
			{
				if (accelerate & MK_SHIFT)
				{
					if (zDelta < 0) scroll_offset += 128 / shader_scale;
					else if (zDelta > 0) scroll_offset -= 128 / shader_scale;
				}
				else
				{
					if (zDelta < 0) scroll_offset += 8 / shader_scale;
					else if (zDelta > 0) scroll_offset -= 8 / shader_scale;
				}
			}
			else//scroll vertical
			{
				if (accelerate & MK_SHIFT)
				{
					if (zDelta < 0) scroll_offset += 128*buffer_width / shader_scale;
					else if (zDelta > 0) scroll_offset -= 128*buffer_width / shader_scale;
				}
				else
				{
					if (zDelta < 0) scroll_offset += 8*buffer_width / shader_scale;
					else if (zDelta > 0) scroll_offset -= 8*buffer_width / shader_scale;
				}
			}
			
			int max = max_file - buffer_width * buffer_height;
			if (scroll_offset < 0) scroll_offset = 0;
			if (scroll_offset > max) scroll_offset = max;
			
			UpdateTitle(hWnd);
		}
		break;
    }
	
    return DefWindowProc (hWnd, message, wParam, lParam);
}


// this function initializes and prepares Direct3D for use
bool initD3D(HWND hWnd)
{
	//initialize specified version of Direct X
    d3d = Direct3DCreate9(D3D_SDK_VERSION);
	
	//check Direct X capabilities
	D3DCAPS9 dx_caps;
	ZeroMemory(&dx_caps, sizeof(D3DCAPS9)); // Can't hurt to zero the struct.
	
	HRESULT hRes = d3d->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &dx_caps);
	
	if (hRes != S_OK || dx_caps.PixelShaderVersion < D3DPS_VERSION(3, 0)) 
	{
		MessageBox(hWnd, "Error initializing DirectX\nShader Model 3.0 Support required\nApplication will close", "Error", MB_OK);
		return false;
	}
	
	//init Direct X
    D3DPRESENT_PARAMETERS d3dpp;
	
    ZeroMemory(&d3dpp, sizeof(d3dpp));
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.hDeviceWindow = hWnd;
    d3dpp.BackBufferFormat = D3DFMT_X8R8G8B8;
    d3dpp.BackBufferWidth = MAX_BUFFER_WIDTH;
    d3dpp.BackBufferHeight = MAX_BUFFER_HEIGHT;
	
    hRes = d3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &d3ddev);
	if (hRes != S_OK)
	{
		MessageBox(hWnd, "Error initializing DirectX\nCreateDevice Failed\nApplication will close", "Error", MB_OK);
		return false;
	}
	
	//load data display effect
	IDirect3DPixelShader9* lpPixelShader = NULL;
	LPD3DXBUFFER pCode = NULL;
	LPD3DXBUFFER pErrorMsgs = NULL;
	
	LPCSTR shader_path = ".\\DataDisplay.fx";
	hRes = D3DXCreateEffectFromFile(d3ddev, shader_path, NULL, NULL, 0, NULL, &dxEffect, &pErrorMsgs);
	if (hRes != S_OK)
	{
		MessageBox(hWnd, "Error initializing DirectX\nUnable to compile or find \".\\DataDisplay.fx\"\nApplication will close", "Error", MB_OK);
		return false;
	}
	
	hRes = dxEffect->SetTechnique("Simplest");
	if (hRes != S_OK)
	{
		MessageBox(hWnd, "Error initializing DirectX\nUnable to set technique to Simplest\nApplication will close", "Error", MB_OK);
		return false;
	}
	
	//load exe as texture
	hRes = D3DXCreateTexture(d3ddev, texture_size, texture_size, 1, D3DUSAGE_DYNAMIC, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &texture);
	if (hRes != S_OK)
	{
		MessageBox(hWnd, "Error initializing DirectX\nUnable to create 4096x4096 D3DFMT_A8R8G8B8 texture\nApplication will close", "Error", MB_OK);
		return false;
	}
	
	HMODULE hModule = GetModuleHandleW(NULL);
	char exePath[MAX_PATH];
	GetModuleFileName(hModule, exePath, MAX_PATH);
	
	bool bRes;
	bRes = write_file_to_texture(exePath, texture);
	if (bRes == false)
	{
		MessageBox(hWnd, "Error initializing DirectX\nUnable to read this executable into texture\nApplication will close", "Error", MB_OK);
		return false;
	}
	dxEffect->SetTexture ("DataTexture", texture);
	
    directx_initialized = init_graphics();
	return directx_initialized;
}

bool write_file_to_texture(const char *file_path, LPDIRECT3DTEXTURE9 texture)
{
	FILE *fp;
	int err = fopen_s(&fp, file_path, "rb");
	if (fp == NULL || err != 0) 
	{
		char a[1024];
		sprintf_s<1024>(a, "\"%s\"\nError loading file\nAttempting to continue", file_path);
		MessageBox(GetForegroundWindow(), a, "Error", MB_OK);
		return false;
	}
	
	//get file size
	fseek(fp, 0L, SEEK_END);
	long size = ftell(fp);
	if (size > max_file)
	{
		MessageBox(GetForegroundWindow(), "Warning file is larger than 64MB\nOnly first 64MB will be loaded", "Warning", MB_OK);
	}
	fseek(fp, 0L, SEEK_SET);
	
	size_t block_size = -1;
	unsigned char *block_data = new unsigned char[texture_size];
	
    D3DLOCKED_RECT lock_rect;
    texture->LockRect(0, &lock_rect, NULL, D3DLOCK_DISCARD);
	//look into why it needs a lock rect
	
	unsigned char *raw_texture_data = (unsigned char *)lock_rect.pBits;
	ZeroMemory(lock_rect.pBits, max_file);
	
	data_array_length = 0;
	scroll_offset = 0;
	while (block_size != 0 && data_array_length < max_file)//note since texture_size if a factor of max_file there is no danger of overrunning the buffer
	{
		block_size = fread(block_data, sizeof(byte), texture_size, fp);
		if (block_size == 0) break;
		
		//copy to offset 
		//write to texture
		memcpy(raw_texture_data, block_data, block_size);
		raw_texture_data += block_size;
		data_array_length += block_size;
	}
	
    texture->UnlockRect(0);
	
	fclose(fp);
	
	delete block_data;
	
	return true;
}

// this is the function used to render a single frame
void render_frame(void)
{
    d3ddev->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
	
    d3ddev->BeginScene();
	
	dxEffect->SetFloat( "Width", (float)buffer_width );
	dxEffect->SetFloat( "Height", (float)buffer_height );
	dxEffect->SetInt  ( "Length", data_array_length );
	dxEffect->SetInt  ( "Scale", shader_scale );
	dxEffect->SetInt  ( "DisplayColor", shader_color );
	dxEffect->SetInt  ( "ScrollOffset", scroll_offset );
	
	UINT uPasses;
	dxEffect->Begin(&uPasses, 0);
	dxEffect->BeginPass(0);
	
    d3ddev->SetFVF(CUSTOMFVF);
    d3ddev->SetStreamSource(0, v_buffer, 0, sizeof(CUSTOMVERTEX));
    d3ddev->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
	
	dxEffect->EndPass();
	dxEffect->End();
	
    d3ddev->EndScene();
	
    d3ddev->Present(NULL, NULL, NULL, NULL);
}


// Free directX memory
void cleanD3D(void)
{
	dxEffect->Release();
    v_buffer->Release();
    d3ddev->Release();
    d3d->Release();
}


// this is the function that puts the 3D models into video RAM
bool init_graphics(void)
{
    // create the vertices using the CUSTOMVERTEX struct
    CUSTOMVERTEX vertices[] =
    {
        { 0.0f, 0.0f, 0.5f, 1.0f, D3DCOLOR_XRGB(0, 0, 255), 0.0f, 0.0f, },
        { MAX_BUFFER_WIDTH, 0.0f, 0.5f, 1.0f, D3DCOLOR_XRGB(0, 255, 0), 1.0f, 0.0f, },
        { 0.0f, MAX_BUFFER_HEIGHT, 0.5f, 1.0f, D3DCOLOR_XRGB(255, 0, 0), 0.0f, 1.0f, },
        { MAX_BUFFER_WIDTH, MAX_BUFFER_HEIGHT, 0.5f, 1.0f, D3DCOLOR_XRGB(255, 255, 255), 1.0f, 1.0f, },
    };
	
    // create a vertex buffer interface called v_buffer
    d3ddev->CreateVertexBuffer(4*sizeof(CUSTOMVERTEX), 0, CUSTOMFVF, D3DPOOL_MANAGED, &v_buffer, NULL);
	
    VOID* pVoid;    // a void pointer
	
    // lock v_buffer and load the vertices into it
    v_buffer->Lock(0, 0, (void**)&pVoid, 0);
    memcpy(pVoid, vertices, sizeof(vertices));
    v_buffer->Unlock();
	
	return true;
}
